import re
from typing import List, Union

import torch

import numpy as np  # type:ignore

from loguru import logger

from ..ml.sbd import load_model
from ..metrics import rouge1_score
from .util import tr_lower, select_layer, __tr_lower_abbrv__, flatten, pad
from .word_tokenizer import get_default_word_tokenizer, WordTokenizer


class Span:
    def __init__(self, i: int, span, doc):
        self.doc = doc
        self.i = i
        self.value = span

    def __str__(self):
        return self.text

    def __repr__(self):
        return self.text

    @property
    def text(self):
        return self.doc.raw[slice(*self.value)]

    def span_features(self):
        is_first_span = self.i == 0
        is_last_span = self.i == len(self.doc.spans) - 1
        word = self.doc.raw[slice(*self.value)]

        if not is_first_span:
            word_m1 = self.doc.raw[slice(*self.doc.spans[self.i - 1].value)]
        else:
            word_m1 = '<D>'

        if not is_last_span:
            word_p1 = self.doc.raw[slice(*self.doc.spans[self.i + 1].value)]
        else:
            word_p1 = '</D>'

        features = {}

        # All upper features
        if word_m1.isupper() and not is_first_span:
            features["PREV_ALL_CAPS"] = True

        if word.isupper():
            features["ALL_CAPS"] = True

        if word_p1.isupper() and not is_last_span:
            features["NEXT_ALL_CAPS"] = True

        # All lower features
        if word_m1.islower() and not is_first_span:
            features["PREV_ALL_LOWER"] = True

        if word.islower():
            features["ALL_LOWER"] = True

        if word_p1.islower() and not is_last_span:
            features["NEXT_ALL_LOWER"] = True

        # Century
        if tr_lower(word_p1).startswith("yüzyıl") and not is_last_span:
            features["NEXT_WORD_CENTURY"] = True

        # Number
        if (tr_lower(word_p1).startswith("milyar") or tr_lower(word_p1).startswith("milyon") or tr_lower(
                word_p1).startswith("bin")) and not is_last_span:
            features["NEXT_WORD_NUMBER"] = True

        # Percentage
        if tr_lower(word_m1).startswith("yüzde") and not is_first_span:
            features["PREV_WORD_PERCENTAGE"] = True

        # In parenthesis feature
        if word[0] == '(' and word[-1] == ')':
            features["IN_PARENTHESIS"] = True

        # Suffix features
        m = re.search(r'\W+$', word)

        if m:
            subspan = m.span()
            suff2 = word[slice(*subspan)][:2]

            if all((c in '!\').?’”":;…') for c in suff2):
                features['NON-ALNUM SUFFIX2'] = suff2

        if word_p1[0] in '-*◊' and not is_last_span:
            features["NEXT_BULLETIN"] = True

        # title features
        # if word_m1.istitle() and not is_first_span:
        #     features["PREV_TITLE"] = 1

        # NOTE: 'Beşiktaş’ın' is not title by python :/
        m = re.search(r'\w+', word)

        if m:
            subspan = m.span()
            if word[slice(*subspan)].istitle():
                features["TITLE"] = True

        m = re.search(r'\w+', word_p1)

        if m:
            subspan = m.span()

            if word_p1[slice(*subspan)].istitle() and not is_last_span:
                features["NEXT_TITLE"] = True

        # suffix features
        for name, symbol in zip(['DOT', 'ELLIPSES', 'ELLIPSES', 'EXCLAMATION', 'QUESTION'],
                                ['.', '...', '…', '?', '!']):
            if word.endswith(symbol):
                features[f"ENDS_WITH_{name}"] = True

        # prefix abbreviation features
        if '.' in word:

            if any(tr_lower(word).startswith(abbrv) for abbrv in __tr_lower_abbrv__):
                features["PREFIX_IS_ABBRV"] = True
            if any(tr_lower(word).endswith(abbrv) for abbrv in __tr_lower_abbrv__):
                features["SUFFIX_IS_ABBRV"] = True
            else:
                prefix = word.split('.', maxsplit=1)[0]
                features["PREFIX_IS_DIGIT"] = prefix.isdigit()

        # if '.' in word_m1:
        #     prefix_m1 = word_m1.split('.', maxsplit=1)[0]
        #
        #     if tr_lower(prefix_m1) in __tr_lower_abbrv__ and not is_first_span:
        #         features[f"PREV_PREFIX_IS_ABBRV"] = True
        #     else:
        #         features["PREV_PREFIX_IS_DIGIT"] = prefix_m1.isdigit()
        #
        # if '.' in word_p1:
        #     prefix_p1 = word_p1.split('.', maxsplit=1)[0]
        #
        #     if tr_lower(prefix_p1) in __tr_lower_abbrv__ and not is_last_span:
        #         features[f"NEXT_PREFIX_IS_ABBRV"] = True
        #     else:
        #         features["NEXT_PREFIX_IS_DIGIT"] = prefix_p1.isdigit()

        return features


class Sentences:
    tokenizer = get_default_word_tokenizer()

    def __init__(self, id_: int, text: str, doc):
        self.id = id_
        self.text = text

        self._tokens = None
        self.document = doc
        self._bert = None
        self.toks = None

    @staticmethod
    def set_word_tokenizer(tokenizer_name):
        if tokenizer_name != Sentences.tokenizer.__name__:
            Sentences.tokenizer = WordTokenizer.factory(tokenizer_name)

    @property
    def bert(self):
        return self._bert

    @bert.setter
    def bert(self, bert):
        self._bert = bert

    @property
    def input_ids(self):
        return Sentences.tokenizer.convert_tokens_to_ids(self.tokens_with_special_symbols)

    @property
    def tokens(self):
        if self._tokens is None:
            self._tokens = Sentences.tokenizer(self.text)

        return self._tokens

    @property
    def tokens_with_special_symbols(self):
        return ['[CLS]'] + self.tokens + ['[SEP]']

    def rouge1(self, metric):
        return rouge1_score(
            flatten([[tr_lower(token) for token in sent.tokens] for sent in self.document.sents if sent.id != self.id]),
            [tr_lower(t) for t in self.tokens], metric)

    def __str__(self):
        return self.text

    def __repr__(self):
        return self.text

    def __len__(self):
        return len(self.tokens)

    def __eq__(self, s: str):
        return self.text == s  # no need for type checking, will return false for non-strings


class Doc:
    sbd = None
    model = None

    def __init__(self, raw: Union[str, None]):
        if Doc.sbd is None and raw is not None:
            logger.info("Loading ML based SBD")
            Doc.sbd = load_model()

        self.raw = raw
        self._bert = None
        self.sents = []
        self.spans = None

        if raw is not None:
            _spans = [match.span() for match in re.finditer(r"\S+", self.raw)]

            self.spans = [Span(i, span, self) for i, span in enumerate(_spans)]

            y_pred = Doc.sbd.predict((span.span_features() for span in self.spans))

            eos_list = [end for (start, end), y in zip(_spans, y_pred) if y == 1]

            if len(eos_list) > 0:
                for i, eos in enumerate(eos_list):
                    if i == 0:
                        self.sents.append(Sentences(i, self.raw[:eos].strip(), self))
                    else:
                        self.sents.append(Sentences(i, self.raw[eos_list[i - 1] + 1:eos].strip(), self))
            else:
                self.sents.append(Sentences(0, self.raw.strip(), self))

    @classmethod
    def from_sentences(cls, sentences: List[str]):

        d = Doc(None)

        for i, s in enumerate(sentences):
            d.sents.append(Sentences(i, s, d))

        d.raw = "\n".join(sentences)

        return d

    def __str__(self):
        return self.raw

    def __repr__(self):
        return self.raw

    def __len__(self):
        return len(self.sents)

    def max_length(self):
        """Maximum length of a sentence including special symbols."""
        return max(len(s.tokens_with_special_symbols) for s in self.sents)

    def padded_matrix(self, return_numpy=False, return_mask=True):
        """Returns a 0 padded numpy.array or torch.tensor
              One row for each sentence
              One column for each token (pad 0 if length of sentence is shorter than the max length)

        :param return_numpy: Whether to return numpy.array or torch.tensor
        :param return_mask: Whether to return padding mask
        :return:
        """
        max_len = self.max_length()

        if not return_numpy:
            mat = torch.tensor([pad(s.input_ids, max_len) for s in self.sents])

            if return_mask:
                return mat, (mat > 0).to(int)
            else:
                return mat
        else:
            mat = np.array([pad(s.input_ids, max_len) for s in self.sents])

            if return_mask:
                return mat, (mat > 0).astype(int)
            else:
                return mat

    @property
    def bert_embeddings(self):
        if self._bert is None:
            inp, mask = self.padded_matrix()

            if Doc.model is None:
                logger.info("Loading BertModel")
                from transformers import BertModel

                Doc.model = BertModel.from_pretrained("dbmdz/bert-base-turkish-cased", output_hidden_states=True)
                Doc.model.eval()

            with torch.no_grad():
                outputs = Doc.model(inp, mask)

            twelve_layers = outputs[2][1:]

            self._bert = select_layer(twelve_layers, [11], return_cls=False)

        return self._bert



from typing import List
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity
import networkx as nx
from ._base import ExtractiveSummarizer
from ..bblock import Sentences
from ..config import tokenizer_context


class TextRank(ExtractiveSummarizer):
    """Assign a higher importance score to longer sentences.

    mode : {'token', 'char'}, default='token'
        Whether the distance should be defined by total number of characters or
        total number of tokens in a sentence.

    normalize : bool, optional (default=True)
        If ``False``, return a raw score vector.
        Otherwise, return a score vector that adds up to 1.
    """

    tags = ExtractiveSummarizer.tags + ['ml', 'rank']

    def __init__(self, input_type="bert", alpha=0.85, normalize=True):
        super().__init__(normalize)

        if input_type not in ['bert']:
            raise ValueError(f"mode should be one of 'bert'")

        self.input_type = input_type
        self.alpha = alpha

    def _predict(self, sentences: List[Sentences]) -> np.ndarray:
        with tokenizer_context('bert', warning=True):
            if len(sentences) == 0:
                raise ValueError(f"Ensure that document contains a few sentences for summarization")

            doc = sentences[0].document

            vectors = doc.bert_embeddings

            similarity_matrix = np.zeros((len(vectors), len(vectors)))

            for i in range(len(vectors)):
                for j in range(len(vectors)):
                    if i == j:
                        continue

                    similarity_matrix[i][j] = cosine_similarity(vectors[i].reshape(1, -1), vectors[j].reshape(1, -1))[
                        0, 0]

            nx_graph = nx.from_numpy_array(similarity_matrix)
            scores = nx.pagerank(nx_graph, alpha=self.alpha)

            return np.array([scores[i] for i in range(len(scores))])



from collections import defaultdict
from math import ceil
from typing import List
from loguru import logger

import click
from tabulate import tabulate
import warnings
import numpy as np  # type: ignore
from sklearn.metrics import ndcg_score  # type: ignore

from sadedegel.dataset import load_annotated_corpus
from sadedegel.summarize import RandomSummarizer, PositionSummarizer, Rouge1Summarizer, KMeansSummarizer, \
    AutoKMeansSummarizer, \
    DecomposedKMeansSummarizer, LengthSummarizer, TextRank
from sadedegel import Sentences, Doc
from sadedegel import tokenizer_context

logger.disable("sadedegel")

SUMMARIZERS = [('Random Summarizer', RandomSummarizer()), ('FirstK Summarizer', PositionSummarizer()),
               ('LastK Summarizer', PositionSummarizer('last')), ('Rouge1 Summarizer (f1)', Rouge1Summarizer()),
               ('Rouge1 Summarizer (precision)', Rouge1Summarizer('precision')),
               ('Rouge1 Summarizer (recall)', Rouge1Summarizer('recall')),
               ('Length Summarizer (char)', LengthSummarizer('token')),
               ('Length Summarizer (token)', LengthSummarizer('char')),
               ('KMeans Summarizer', KMeansSummarizer()),
               ('AutoKMeans Summarizer', AutoKMeansSummarizer()),
               ('DecomposedKMeans Summarizer', DecomposedKMeansSummarizer()),
               ("TextRank Summarizer (BERT)", TextRank())]


def to_sentence_list(sents: List[str]) -> List[Sentences]:
    l: List[Sentences] = []

    for i, sent in enumerate(sents):
        l.append(Sentences(i, sent, l))

    return l


@click.group(help="SadedeGel summarizer commandline")
def cli():
    pass


def dot_progress(i, length):
    n = ceil(length * 0.05)

    if i == length - 1:
        click.echo(".", nl=True, color="yellow")
    elif i % n == 0 and i != 0:
        click.echo(".", nl=False, color="yellow")


@cli.command()
@click.option("-f", "--table-format", default="github")
@click.option("-t", "--tag", default=["extractive"], multiple=True)
@click.option("-d", "--debug", default=False)
def evaluate(table_format, tag, debug):
    """Evaluate all summarizers in sadedeGel"""
    if not debug:
        warnings.filterwarnings("ignore")

    anno = load_annotated_corpus(False)
    relevance = [[doc['relevance']] for doc in anno]

    summarizers = [summ for summ in SUMMARIZERS if any(_tag in summ[1] for _tag in tag)]

    scores = defaultdict(list)

    for word_tokenizer in ['simple', 'bert']:
        click.echo("Word Tokenizer: " + click.style(f"{word_tokenizer}", fg="blue"))
        docs = [Doc.from_sentences(doc['sentences']) for doc in anno]  # Reset document because of memoization
        with tokenizer_context(word_tokenizer):
            for name, summarizer in summarizers:
                click.echo(click.style(f"    {name} ", fg="magenta"), nl=False)
                # skip simple tokenizer for clustering models
                if ("cluster" in summarizer or "rank" in summarizer) and word_tokenizer == "simple":
                    click.echo(click.style("SKIP", fg="yellow"))
                    continue

                for i, (y_true, d) in enumerate(zip(relevance, docs)):
                    dot_progress(i, len(relevance))

                    y_pred = [summarizer.predict(d.sents)]

                    score_10 = ndcg_score(y_true, y_pred, k=ceil(len(d) * 0.1))
                    score_50 = ndcg_score(y_true, y_pred, k=ceil(len(d) * 0.5))
                    score_80 = ndcg_score(y_true, y_pred, k=ceil(len(d) * 0.8))

                    scores[f"{name} - {word_tokenizer}"].append((score_10, score_50, score_80))

    table = [[algo, np.array([s[0] for s in scores]).mean(), np.array([s[1] for s in scores]).mean(),
              np.array([s[2] for s in scores]).mean()] for
             algo, scores in scores.items()]

    # TODO: Sample weight of instances.
    print(
        tabulate(table, headers=['Method & Tokenizer', 'ndcg(k=0.1)', 'ndcg(k=0.5)', 'ndcg(k=0.8)'],
                 tablefmt=table_format,
                 floatfmt=".4f"))

    if debug:
        click.echo(np.array(table).shape)


if __name__ == '__main__':
    cli()



from collections import Counter
from typing import List
import warnings

_AVAILABLE_METRICS = ["f1", "recall", "precision"]


def _get_overlap_count(y_ref: List, y_cand: List) -> int:
    counter_ref = Counter(y_ref)
    counter_cand = Counter(y_cand)

    return sum((counter_ref & counter_cand).values())


def _get_recall(y_ref: list, y_cand: list) -> float:
    if len(y_ref) == 0:
        warnings.warn("y_ref is empty causing division by zero", UserWarning)
        return 0
    else:
        overlap_count = _get_overlap_count(y_ref, y_cand)
        return overlap_count / len(y_ref)


def _get_precision(y_ref: list, y_cand: list) -> float:
    if len(y_cand) == 0:
        warnings.warn("y_ref is empty causing division by zero", UserWarning)
        return 0
    else:
        overlap_count = _get_overlap_count(y_ref, y_cand)
        return overlap_count / len(y_cand)


def _get_f1(y_ref: list, y_cand: list) -> float:
    recall = _get_recall(y_ref, y_cand)
    precision = _get_precision(y_ref, y_cand)

    if recall == 0 and precision == 0:
        warnings.warn("Both precision & recall is 0 causing division by zero in evaluation f1-score", UserWarning)
        f1 = 0.
    else:
        f1 = (2 * precision * recall) / (precision + recall)

    return f1


def rouge1_score(y_ref: List, y_cand: List, metric: str = "f1"):
    if metric.lower() not in _AVAILABLE_METRICS:
        raise ValueError(f"metrics ({metric}) should be one of {_AVAILABLE_METRICS}")

    if not (type(y_ref) == list and type(y_cand) == list):
        raise ValueError("Both inputs (y_ref & y_cand) should be of list type.")

    if metric == "recall":
        return _get_recall(y_ref, y_cand)
    elif metric == "precision":
        return _get_precision(y_ref, y_cand)
    else:
        return _get_f1(y_ref, y_cand)



import click
import os.path
from loguru import logger
from pathlib import Path
import boto3
from smart_open import open
import tarfile
import glob
from tqdm import tqdm
from sadedegel.tokenize import Doc
import json
import sys

from ..util import safe_read
from ._core import load_extended_metadata


@click.group(help="Extended Dataset commandline")
def cli():
    pass


@cli.command()
@click.option("--access-key", help="Access Key ID to download dataset.", prompt=True,
              default=lambda: os.environ.get('sadedegel_access_key', ''))
@click.option("--secret-key", help="Secret Key ID to download dataset.", prompt=True,
              default=lambda: os.environ.get('sadedegel_secret_key', ''))
@click.option("--data_home", '-d', help="Data home directory", default="~/.sadedegel_data")
def download(access_key, secret_key, data_home):
    """Download extended corpus from cloud with your key."""

    data_home = Path(os.path.expanduser(data_home)) / 'extended'
    logger.info(f"Data directory for extended data {data_home}")

    data_home.mkdir(parents=True, exist_ok=True)

    transport_params = {
        'session': boto3.Session(aws_access_key_id=access_key,
                                 aws_secret_access_key=secret_key),
        'resource_kwargs': {
            'endpoint_url': 'https://storage.googleapis.com',
        }
    }

    __tarballs__ = ['cumhuriyet.tar.gz', 'haberturk.tar.gz', 'hurriyet.tar.gz', 'milliyet.tar.gz',
                    'yetkin-report.tar.gz']

    click.secho("Downloading corpus...")

    for tarball in __tarballs__:
        url = f"s3://sadedegel/dataset/raw/{tarball}"

        with open(url, 'rb', transport_params=transport_params) as fp:
            tf = tarfile.open(fileobj=fp)

            tf.extractall(data_home.absolute())

        url = f"s3://sadedegel/dataset/sents/{tarball}"

        with open(url, 'rb', transport_params=transport_params) as fp:
            tf = tarfile.open(fileobj=fp)

            tf.extractall(data_home.absolute())

        click.secho(f".done: {tarball}", fg="green")


@cli.command()
@click.option("--data_home", '-d', help="Data home directory", default="~/.sadedegel_data")
def sbd(data_home):
    """Generate sentence boundary detected corpus out of raw document corpus."""

    data_home = Path(os.path.expanduser(data_home))

    logger.info(f"Data directory for extended data {data_home}")

    raw_dir = data_home / 'extended' / 'raw'

    for section in raw_dir.iterdir():

        sents_dir = section.parent.parent / 'sents' / str(section.name)

        if section.is_dir():
            sents_dir.mkdir(parents=True, exist_ok=True)

            for raw in tqdm(glob.glob(str((raw_dir / section / '*.txt').absolute())), desc=f"{section.name} documents",
                            unit="doc"):
                fn_noext, _ = os.path.splitext(os.path.basename(raw))

                target = (sents_dir / f"{fn_noext}.json").absolute()

                if not os.path.exists(target) or (os.path.exists(target) and os.path.getsize(target) == 0):
                    try:
                        d = Doc(safe_read(raw))

                        with open(target, 'w') as wp:
                            json.dump(
                                dict(sentences=[s.text for s in d.sents], rouge1=[s.rouge1("f1") for s in d.sents]),
                                wp,
                                ensure_ascii=False)
                    except:
                        logger.exception(f"Error in processing document {raw}")

                        raise


@cli.command()
@click.option("--data_home", '-d', help="Data home directory", default="~/.sadedegel_data")
def metadata(data_home):
    """Check extended dataset corpus metadata"""
    from pprint import pformat

    md = load_extended_metadata(data_home)

    if md is not None:
        click.secho(pformat(md), fg="yellow")


if __name__ == '__main__':
    cli()



import warnings
import json


def safe_read(file: str):
    try:
        with open(file) as fp:
            return fp.read()
    except:
        warnings.warn(f"Error in reading {file}", UserWarning)
        raise


def safe_json_load(file: str):
    try:
        return json.loads(safe_read(file))
    except:
        warnings.warn(f"JSON Decoding error for {file}", UserWarning)
        raise



from os.path import dirname
from pathlib import Path

from sklearn.ensemble import RandomForestClassifier  # type: ignore
from sklearn.feature_extraction import DictVectorizer  # type: ignore
from sklearn.pipeline import Pipeline  # type: ignore
from joblib import dump, load  # type: ignore

from loguru import logger


def create_model():
    """Creates a new sbd model detector."""
    return Pipeline([('feat', DictVectorizer()), ('dt', RandomForestClassifier())])


def save_model(model, name="sbd.pickle"):
    model_file = (Path(dirname(__file__)) / 'model' / name).absolute()

    dump(model, model_file)


def load_model(name="sbd.pickle"):
    model_file = (Path(dirname(__file__)) / 'model' / name).absolute()
    logger.info(f"Loading sbd model from {model_file}")

    return load(model_file)



from typing import List
import numpy as np
from ._base import ExtractiveSummarizer
from ..bblock import Sentences


class RandomSummarizer(ExtractiveSummarizer):
    """Assign a random importance score to each sentences.

    seed: int, optional(default=42)
        Random generator seed.

    normalize : bool, optional (default=True)
        If ``False``, return a raw score vector.
        Otherwise, return a score vector that adds up to 1.
    """

    tags = ExtractiveSummarizer.tags + ['baseline']

    def __init__(self, seed=42, normalize=True):
        super().__init__(normalize)
        self.seed = seed
        np.random.seed(self.seed)

    def _predict(self, sentences: List[Sentences]) -> np.ndarray:
        if type(sentences) == list:
            n = len(sentences)
        else:
            n = len(list(sentences))

        return np.random.random(n)


class LengthSummarizer(ExtractiveSummarizer):
    """Assign a higher importance score to longer sentences.

    mode : {'token', 'char'}, default='token'
        Whether the distance should be defined by total number of characters or
        total number of tokens in a sentence.

    normalize : bool, optional (default=True)
        If ``False``, return a raw score vector.
        Otherwise, return a score vector that adds up to 1.
    """

    tags = ExtractiveSummarizer.tags + ['baseline']

    def __init__(self, mode="token", normalize=True):
        super().__init__(normalize)

        if mode not in ['token', 'char']:
            raise ValueError(f"mode should be one of 'token', 'char'")

        self.mode = mode

    def _predict(self, sentences: List[Sentences]) -> np.ndarray:

        if self.mode == 'token':
            scores = np.array([len(sent.tokens) for sent in sentences])
        else:
            scores = np.array([sum(len(token) for token in sent.tokens) for sent in sentences])

        return scores


class PositionSummarizer(ExtractiveSummarizer):
    """Assign a higher importance score based on relative position in documentation.

    mode : {'first', 'last'}, default='first'
        Whether the smaller position indices have higher scores.

    normalize : bool, optional (default=True)
        If ``False``, return a raw score vector.
        Otherwise, return a score vector that adds up to 1.
    """

    tags = ExtractiveSummarizer.tags + ['baseline']

    def __init__(self, mode='first', normalize=True):
        super().__init__(normalize)

        if mode not in ['first', 'last']:
            raise ValueError(f"mode should be one of 'first', 'last'")

        self.mode = mode

    def _predict(self, sentences: List[Sentences]) -> np.ndarray:
        if type(sentences) == list:
            n = len(sentences)
        else:
            n = len(list(sentences))

        if self.mode == "first":
            scores = np.arange(n - 1, -1, -1)
        else:
            scores = np.arange(n)

        return scores


class BandSummarizer(ExtractiveSummarizer):
    """Split document into bands of length k.

    Scoring scheme has two parts:
        1. Each band is scored by PositionSummarizer individually.
        2. Relative score for the same relative position within the band is also define by `mode` parameter.

    mode : {'first', 'last'}, default='first'
        Whether the smaller position indices within a band have higher scores.

    k : int, default=3
        Split document into bands of length k sentences (except the last one).

    normalize : bool, optional (default=True)
        If ``False``, return a raw score vector.
        Otherwise, return a score vector that adds up to 1.
    """

    tags = ExtractiveSummarizer.tags + ['baseline']

    def __init__(self, mode='first', k=3, normalize=True):
        super().__init__(normalize)
        self.k = k

        if mode not in ['first', 'last']:
            raise ValueError(f"mode should be one of 'first', 'last'")

        self.mode = mode

    def _predict(self, sentences: List[Sentences]) -> np.ndarray:
        raise NotImplementedError("BandSummarizer is not completed yet.")



from typing import Tuple
import click
import requests
from .about import __version__, __herokuapp_url__


def to_tuple(version_string: str) -> Tuple:
    return tuple(version_string.split('.'))


def to_str(version_tuple: Tuple) -> str:
    return ".".join(version_tuple)


@click.group(help="SadedeGel commandline")
def cli():
    pass


@cli.command()
def info():
    """SadedeGel version information in details"""
    most_recent_version = requests.get("https://pypi.python.org/pypi/sadedegel/json").json()['info']['version']
    most_recent_version = to_tuple(most_recent_version)

    click.echo(f"sadedeGel (@PyPI): {click.style(to_str(most_recent_version), fg='green')}")

    installed_version = to_tuple(__version__)

    if installed_version < most_recent_version:
        color = "yellow"
    else:
        color = "green"

    click.echo(f"sadedeGel (installed): {click.style(to_str(installed_version), fg=color)}")

    heroku_version = requests.get(f"{__herokuapp_url__}/api/info").json()['version']
    heroku_version = to_tuple(heroku_version)

    if heroku_version < most_recent_version:
        color = "yellow"
    else:
        color = "green"

    click.echo(f"sadedeGel Server ({__herokuapp_url__}): {click.style(to_str(heroku_version), fg=color)}")


if __name__ == '__main__':
    cli()  # pylint: disable=E1120



__all__ = ['set_config', 'get_config', 'describe_config', 'get_all_configs']

from typing import Any
from functools import wraps
from collections import namedtuple
from contextlib import contextmanager
import warnings
from .bblock.doc import Sentences

Configuration = namedtuple("Configuration", "config, description, valid_values")

configs = {
    "word_tokenizer": Configuration(config="word_tokenizer",
                                    description="Change the default word tokenizer used by sadedegel",
                                    valid_values=None)
}


def check_config(f):
    @wraps(f)
    def wrapper(*args, **kwds):
        config = args[0]
        if config not in configs:
            raise Exception((f"{config} is not a valid configuration for sadegel."
                             "Use sadedegel.get_all_configs() to access list of valid configurations."))
        return f(*args, **kwds)

    return wrapper


def check_value(f):
    @wraps(f)
    def wrapper(*args, **kwds):
        config, value = args[0], args[1]
        cfg = configs.get(config, None)

        if cfg:
            if value not in cfg.valid_values:
                raise Exception(
                    f"{value} is not a valid value for {config}. Choose one of {', '.join(cfg.valid_values)}")
        else:
            raise Exception((f"{config} is not a valid configuration for sadegel."
                             "Use sadedegel.get_all_configs() to access list of valid configurations."))

        return f(*args, **kwds)

    return wrapper


@check_config
def set_config(config: str, value: Any):
    if config == "word_tokenizer":
        Sentences.set_word_tokenizer(value)


@contextmanager
def tokenizer_context(tokenizer_name, warning=False):
    current = Sentences.tokenizer.__name__

    if warning and current != tokenizer_name:
        warnings.warn(f"Changing tokenizer to {tokenizer_name}")

    try:
        set_config("word_tokenizer", tokenizer_name)
        yield
    finally:
        set_config("word_tokenizer", current)


@check_config
def get_config(config: str):  # pylint: disable=inconsistent-return-statements
    if config == "word_tokenizer":
        return Sentences.tokenizer.__name__


@check_config
def describe_config(config: str, print_desc=False):  # pylint: disable=inconsistent-return-statements
    if configs[config].valid_values is not None:
        valid_values_fragment = "\n\nValid values are\n" + "\n".join(configs[config].valid_values)
    else:
        valid_values_fragment = ""

    config_doc = f"{configs[config].description}{valid_values_fragment}"

    if print_desc:
        print(config_doc)
    else:
        return config_doc


def get_all_configs():
    return configs



import click
import os.path
from os.path import join as pjoin
import glob
from loguru import logger
import json
from tqdm import tqdm
import sys


@click.group(help="Dataset commandline")
def cli():
    pass


def get_tokenizer(tokenizer: str):
    if tokenizer == 'nltk-punct':
        from ..tokenize import NLTKPunctTokenizer

        return NLTKPunctTokenizer()
    else:
        from ..tokenize import RegexpSentenceTokenizer

        return RegexpSentenceTokenizer()


@cli.command(help="Generate sentences from raw news text docs.")
@click.option('--dataset-dir', help="Dataset directory")
@click.option('--tokenizer', type=click.Choice(['nltk-punct', 're']), default='re',
              help="Sentences tokenizer.")
@click.option('--force', is_flag=True, default=False, help="Overwrite existing json files.")
def sentences(dataset_dir: str, tokenizer: str, force: bool):
    dataset = os.path.dirname(__file__) if dataset_dir is None else dataset_dir

    raws = glob.glob(pjoin(dataset, 'raw', '*.txt'))
    sents_dir = pjoin(dataset, 'sents')

    os.makedirs(sents_dir, exist_ok=True)

    toker = get_tokenizer(tokenizer)
    n = 0

    logger.info("|raw documents|: {}".format(len(raws)))
    logger.info("Sentence tokenizer: {}".format(tokenizer))

    for file in tqdm(raws):
        basename = os.path.basename(file)
        name, _ = os.path.splitext(basename)

        sentences_json_file = pjoin(sents_dir, "{}.json".format(name))

        if not os.path.exists(sentences_json_file) or force:
            with open(sentences_json_file, 'w') as wp, open(file) as fp:
                json.dump(dict(sentences=toker(fp.read())), wp)

            n += 1

    logger.info("Total number of files dumped is {}".format(n))


@cli.command(help="Validate raw & sentences datasets")
@click.option("-v", count=True)
@click.option("--base-path", help="Base data path", default=None)
def validate(v, base_path):
    from sadedegel.dataset import load_raw_corpus, load_sentence_corpus, load_annotated_corpus, file_paths, CorpusTypeEnum

    click.secho("Corpus loading...")
    raw = load_raw_corpus(False, base_path)
    sents = load_sentence_corpus(False, base_path)
    anno = load_annotated_corpus(False, base_path)

    click.secho(".done.", fg="green")
    click.secho(f"Number of News Documents (raw): {len(raw)}".rjust(50))
    click.secho(f"Number of News Documents (sentences): {len(sents)}".rjust(50))
    click.secho(f"Number of News Documents (annotated): {len(anno)}".rjust(50))

    if len(anno) != len(sents):
        anno_files = file_paths(CorpusTypeEnum.ANNOTATED, True, True, base_path)
        sent_files = file_paths(CorpusTypeEnum.SENTENCE, True, True, base_path)

        click.secho("\nSymmetric Difference between sentences & annotated corpus.")

        for diff in set(anno_files).symmetric_difference(set(sent_files)):
            click.secho(f"{diff}".rjust(50))
        click.secho(".warn", fg="yellow")

    click.secho("\nPerforming span checks...")

    for a, b, file in zip(raw, sents, file_paths()):
        for i, sent in enumerate(b['sentences']):
            if sent not in a:
                logger.error(f"""{sent}[{i}] \n\t\t is not a span in raw document \n {a} \n\n Corpus file: {file}
                """)
                sys.exit(1)

    click.secho(".done", fg="green")

    click.secho("\nPerforming span order checks...")

    for a, b, file in zip(raw, sents, file_paths()):

        start = 0
        for i, sent in enumerate(b['sentences']):

            idx = a.find(sent, start)

            if idx == -1:
                logger.error(
                    f"""{sent}[{i}] \n\t\t is potential our of order in "sentences" array of sentence corpus\n {a} \n\n Corpus file: {file}
                    """)
                sys.exit(1)
            else:
                start = start + len(sent)

    click.secho(".done", fg="green")

    click.secho("\nComparing annotated corpus with sentences corpus...")

    anno_names = file_paths(CorpusTypeEnum.ANNOTATED, noext=True, use_basename=True, base_path=base_path)
    sents_names = file_paths(CorpusTypeEnum.SENTENCE, noext=True, use_basename=True, base_path=base_path)

    anno_dict = dict((name, doc) for name, doc in zip(anno_names, anno))
    sents_dict = dict((name, doc) for name, doc in zip(sents_names, sents))

    match = 0

    for _name, _anno in anno_dict.items():
        sent = sents_dict[_name]

        if sent['sentences'] != _anno['sentences']:
            click.secho(f"\nSentences in annotated corpus {_name} doesn't match with document in sentence corpus.")
            sys.exit(1)
        else:
            match += 1

    click.secho(f".done ({match}/{len(anno_dict)})", fg="green")


if __name__ == '__main__':
    cli()



from typing import List, Union
import numpy as np  # type: ignore
from abc import ABC, abstractmethod
from ..bblock import Doc, Sentences


def get_sentences_list(X: Union[Doc, List[Sentences], List[str]]):
    if type(X) == Doc:
        sentences = X.sents  # Get the list of Sentences from Doc
    elif type(X) != list:
        raise ValueError(f"sents parameter should be one of Doc, List[Sentences] or List[str]. Found {type(X)}")
    elif all(type(s) == str for s in X):
        d = Doc.from_sentences(X)
        sentences = d.sents
    elif all(type(s) == Sentences for s in X):
        sentences = X
    else:
        raise ValueError(f"sents parameter should be one of Doc, List[Sentences] or List[str]. Found {type(X)}")

    return sentences


class ExtractiveSummarizer(ABC):
    tags = ['extractive']

    def __init__(self, normalize=True):
        self.normalize = normalize

    @abstractmethod
    def _predict(self, sents: List[Sentences]) -> np.ndarray:
        pass

    def predict(self, sents: Union[Doc, List[Sentences], List[str]]) -> np.ndarray:
        """Predict relevance score for X

        Parameters
        ----------
        sents:
          Doc: sadedeGel Document
          List[Sentences]: List of sadedegel Sentences
          List[str]: List of strings (each element is taken to a sentences)

        Returns
        -------
        Relevance score per sentence
        np.array
        """
        sents = get_sentences_list(sents)

        scores = self._predict(sents)

        if self.normalize:
            return scores / scores.sum()
        else:
            return scores

    def __call__(self, sents: Union[Doc, List[Sentences], List[str]], k: int) -> List[Sentences]:

        sents = get_sentences_list(sents)
        scores = self.predict(sents)

        topk_inds = np.argpartition(scores, k)[-k:]  # returns indices of k top sentences
        topk_inds.sort()  # fix the order of sentences

        summ = [sents[i] for i in topk_inds]

        return summ

    def __contains__(self, tag: str) -> bool:
        """Check whether instance tags contain a given tag for filtering summarizers.
        """

        return tag in self.tags



from enum import Enum

from fastapi import FastAPI
from fastapi.openapi.utils import get_openapi
from fastapi.responses import RedirectResponse
from fastapi.middleware.cors import CORSMiddleware

import uvicorn

from pydantic import BaseModel
from pydantic.typing import List

from sadedegel.bblock import Doc, Sentences
from sadedegel.summarize import RandomSummarizer, PositionSummarizer, Rouge1Summarizer
from sadedegel.about import __version__

import click
import numpy as np

from loguru import logger


class TimeUnitEnum(str, Enum):
    """Time unit is used to specify duration reader wants to spend in reading.
    Together with wpm defines the summary length to be returned.
    """
    SECOND = 'second'
    MINUTE = 'minute'


class BasicRequest(BaseModel):
    """Summarization request."""
    doc: str
    wpm: int = 170


class Request(BasicRequest):
    """Summarization request with duration added."""
    duration: int = 5
    unit: TimeUnitEnum = TimeUnitEnum.MINUTE


class DocSummary(BaseModel):
    """Document statistics summary.
    Used by summarization service calculate front end material.
    """
    sentence_count: int
    word_count: int


class RichSummary(DocSummary):
    """Enhanced statistics summary with reading duration."""
    wpm: int
    duration: float
    unit: TimeUnitEnum


class Response(BaseModel):
    sentences: List[str]
    original: DocSummary
    summary: DocSummary


class APIInfo(BaseModel):
    """API Info including
    Version installed
    Maintainers email
    sadedeGel website
    Current pyPI version available
    """
    version: str
    email: str
    website: str
    github: str


app = FastAPI()


def custom_openapi():
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="sadedeGel",
        version=__version__,
        description="sadedeGel is an extraction-based summarizer for Turkish news content",
        routes=app.routes,
    )
    openapi_schema["info"]["x-logo"] = {
        "url": "https://sadedegel.ai/dist/img/logo-2.png?s=280&v=4"
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi

app.add_middleware(
    CORSMiddleware,
    allow_origins='*',
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


def summary_filter(sents, scores, word_count, limit=None):
    if not (len(sents) == len(scores) and len(scores) == len(word_count)):
        raise ValueError(
            f"Length of sents ({len(sents)}), scores ({len(scores)}) and word_count ({len(word_count)}) should match.")

    rank = np.argsort(scores)[::-1]

    if limit:
        selected_sents_idx = rank[word_count[rank].cumsum() <= limit]
    else:
        selected_sents_idx = rank

    selected_sents_idx.sort()  # sort in ascending order to preserve order

    return sents[selected_sents_idx]


@app.get("/", include_in_schema=False)
async def home():
    return RedirectResponse("http://sadedegel.ai")


@app.get("/api/info", tags=["Information"], summary="sadedeGel metadata")
async def info():
    return APIInfo(version=__version__, email="info@sadedegel.ai", website="sadedegel.ai",
                   github="https://github.com/GlobalMaksimum/sadedegel")


def summarize(summarizer, sentences: List[Sentences], limit: float) -> Response:
    word_count = np.array([len(s) for s in sentences], dtype=np.int)

    scores = summarizer.predict(sentences)

    logger.info(scores)
    logger.info(word_count)

    sentences_limited = summary_filter(np.array(sentences), scores, word_count, limit)

    logger.info(sentences_limited.tolist())

    return Response(sentences=[s.text for s in sentences_limited.tolist()],
                    original=DocSummary(sentence_count=len(sentences), word_count=word_count.sum()),
                    summary=DocSummary(sentence_count=len(sentences_limited),
                                       word_count=np.array([len(s) for s in sentences_limited.tolist()],
                                                           dtype=np.int).sum()))


@app.post("/api/doc/statistics", tags=["Utility"], summary="Calculate approximate reading duration of a document")
async def duration(req: BasicRequest):
    """Calculates the approximate reading duration for the document based on reader read speed"""

    sentences = Doc(req.doc).sents

    word_count = sum((len(s) for s in sentences))

    dur = (word_count / (req.wpm / 60))
    unit = TimeUnitEnum.SECOND

    logger.info(f"Total duration {dur}")

    return RichSummary(sentence_count=len(sentences), word_count=word_count, wpm=req.wpm, duration=dur, unit=unit)


@app.post("/api/summarizer/random", tags=["Summarizer"], summary="Use RandomSummarizer")
async def random(req: Request):
    """Baseline [Random Summarizer](https://github.com/GlobalMaksimum/sadedegel/tree/master/sadedegel/summarize/README.md)

            Picks up random sentences until total number of tokens is less than equal to `wpm x duration`
    """

    sentences = Doc(req.doc).sents

    if req.unit == TimeUnitEnum.MINUTE:
        duration_in_min = req.duration
    else:
        duration_in_min = req.duration / 60.

    return summarize(RandomSummarizer(seed=None), sentences, req.wpm * duration_in_min)


@app.post("/api/summarizer/firstk", tags=["Summarizer"],
          summary="Use Position Summarizer to obtain a FirstK summarizer")
async def firstk(req: Request):
    """Baseline [Position Summarizer](https://github.com/GlobalMaksimum/sadedegel/tree/master/sadedegel/summarize/README.md)

            Picks up first a few sentences until total number of tokens is less than equal to `wpm x duration`
    """

    sentences = Doc(req.doc).sents

    if req.unit == TimeUnitEnum.MINUTE:
        duration_in_min = req.duration
    else:
        duration_in_min = req.duration / 60.

    return summarize(PositionSummarizer(), sentences, req.wpm * duration_in_min)


@app.post("/api/summarizer/rouge1", tags=["Summarizer"],
          summary="Use unsupervised Rouge1 Summarizer")
async def rouge1(req: Request):
    """Baseline [Rouge1 Summarizer](https://github.com/GlobalMaksimum/sadedegel/tree/master/sadedegel/summarize/README.md)

            Rank sentences based on their rouge1 score in Document and return a list of sentences until number of total tokens is less than equal to `wpm x duration`
    """

    sentences = Doc(req.doc).sents

    if req.unit == TimeUnitEnum.MINUTE:
        duration_in_min = req.duration
    else:
        duration_in_min = req.duration / 60.

    return summarize(Rouge1Summarizer(), sentences, req.wpm * duration_in_min)


@click.command()
@click.option("--host", '-h', help="Hostname", default="0.0.0.0")  # nosec
@click.option("--log-level",
              type=click.Choice(['debug', 'info'], case_sensitive=False), help="Logging Level", default="info")
@click.option("--reload", is_flag=True, default=False, help="enable/disable auto reload for development.")
@click.option("--port", help="Port", default=8000)
def server(host, log_level, reload, port):
    """Span a sadedeGel http server instance."""

    if reload:
        uvicorn.run("sadedegel.server.__main__:app", host=host, log_level=log_level, reload=reload, port=port)
    else:
        uvicorn.run(app, host=host, log_level=log_level, reload=reload, port=port)


if __name__ == "__main__":
    server()



from abc import ABC, abstractmethod
from typing import List
from .word_tokenizer_helper import word_tokenize
import warnings


class WordTokenizer(ABC):
    __instances = {}

    @abstractmethod
    def _tokenize(self, text: str) -> List[str]:
        pass

    @abstractmethod
    def convert_tokens_to_ids(self, tokens: List[str]) -> List[int]:
        pass

    def __call__(self, sentence: str) -> List[str]:
        return self._tokenize(str(sentence))

    @staticmethod
    def factory(tokenizer_name: str):
        normalized_name = tokenizer_name.lower().replace(' ', '').replace('-', '').replace('tokenizer', '')
        if normalized_name not in WordTokenizer.__instances:
            if normalized_name == "bert":
                WordTokenizer.__instances[normalized_name] = BertTokenizer()
            elif normalized_name == "simple":
                warnings.warn(
                    ("Note that SimpleTokenizer is pretty new in sadedeGel. "
                     "If you experience any problems, open up a issue "
                     "(https://github.com/GlobalMaksimum/sadedegel/issues/new)"))
                WordTokenizer.__instances[normalized_name] = SimpleTokenizer()
            else:
                raise Exception(
                    (f"No word tokenizer type match with name {tokenizer_name}."
                     " Use one of 'bert-tokenizer', 'SimpleTokenizer', etc."))

        return WordTokenizer.__instances[normalized_name]


class BertTokenizer(WordTokenizer):
    __name__ = "BertTokenizer"

    def convert_tokens_to_ids(self, tokens: List[str]) -> List[int]:
        return self.tokenizer.convert_tokens_to_ids(tokens)

    def __init__(self):
        super(BertTokenizer, self).__init__()

        from transformers import AutoTokenizer
        self.tokenizer = AutoTokenizer.from_pretrained("dbmdz/bert-base-turkish-cased")

    def _tokenize(self, text: str) -> List[str]:
        return self.tokenizer.tokenize(text)


class SimpleTokenizer(WordTokenizer):
    __name__ = "SimpleTokenizer"

    def __init__(self):
        super(SimpleTokenizer, self).__init__()
        self.tokenizer = word_tokenize

    def _tokenize(self, text: str) -> List[str]:
        return self.tokenizer(text)

    def convert_tokens_to_ids(self, ids: List[str]) -> List[int]:
        raise NotImplementedError("convert_tokens_to_ids is not implemented for SimpleTokenizer yet. Use BERTTokenizer")


def get_default_word_tokenizer() -> WordTokenizer:
    return WordTokenizer.factory(BertTokenizer.__name__)



from sadedegel.dataset import load_raw_corpus, load_sentence_corpus, file_paths
from sadedegel.tokenize import NLTKPunctTokenizer, RegexpSentenceTokenizer, Doc
from sadedegel.bblock.util import flatten, is_eos
from sadedegel.ml import create_model, save_model
from sklearn.model_selection import cross_val_score

import click
from typing import List


@click.group(help="sadedeGel SBD commandline")
def tokenize():
    pass


def iou_eval(name: str, y_true: List[List[str]], y_pred: List[List[str]], paths=None):
    i, u, iou = 0, 0, 0

    for idx, (seq1, seq2) in enumerate(zip(y_true, y_pred)):
        s1 = list(map(lambda x: x.strip(), seq1))
        s2 = list(map(lambda x: x.strip(), seq2))

        i += len(set(s1) & set(s2))
        u += len(set(s1) | set(s2))

        delta = len(set(s1) & set(s2)) / len(set(s1) | set(s2))

        if paths:
            click.secho(f"Lost {1.0 - delta:.2f} because of {paths[idx]}", fg="yellow")

        iou += delta

    click.secho(f"IoU score for {name}")

    click.secho(f"Micro IoU: {i / u:.4f}".rjust(25))
    click.secho(f"Macro IoU: {iou / len(y_true):.4f}".rjust(25))


@tokenize.command()
@click.option("-v", count=True, help="verbosity")
def evaluate(v):
    """Evaluate IoU metric for different SBD algorithms over our stock dataset."""
    click.secho("Loading corpus...")
    raw, sents = load_raw_corpus(False), load_sentence_corpus()

    nltk = NLTKPunctTokenizer()
    reg = RegexpSentenceTokenizer()

    y_pred = [nltk(doc) for doc in raw]
    y_true = [doc['sentences'] for doc in sents]

    iou_eval("NLTKPunctTokenizer", y_true, y_pred, file_paths() if v > 0 else None)

    y_pred = [reg(doc) for doc in raw]

    iou_eval("RegexpSentenceTokenizer", y_true, y_pred, file_paths() if v > 0 else None)

    y_pred = [[s.text for s in Doc(doc).sents] for doc in raw]

    iou_eval("MLBasedTokenizer", y_true, y_pred, file_paths() if v > 0 else None)


@tokenize.command()
def diff():
    """Git like diff tool to compare sentence generated by our tokenizer vs actual list of sentences."""
    click.secho("Loading corpus...")
    raw, sents = load_raw_corpus(False), load_sentence_corpus()

    y_true = [doc['sentences'] for doc in sents]

    y_pred = [Doc(doc).sents for doc in raw]

    paths = file_paths()

    for i in range(len(y_true)):

        if y_true[i] != y_pred[i]:
            click.secho(f"Document {paths[i]}")
            for s_true in y_true[i]:
                if s_true not in y_pred[i]:
                    click.secho(f"+ {s_true}", fg="green")

            click.secho()

        for s_pred in y_pred[i]:
            if s_pred not in y_true[i]:
                click.secho(f"- {s_pred}", fg="red")

        click.secho()
        click.secho()


@tokenize.command()
def build():
    """Build a ML based SBD"""

    raw_corpus = load_raw_corpus(False)
    sent_corpus = load_sentence_corpus(False)

    features = flatten([[span.span_features() for span in Doc(raw).spans] for raw in raw_corpus])
    y = flatten(
        [[is_eos(span, sent['sentences']) for span in Doc(raw).spans] for raw, sent in zip(raw_corpus, sent_corpus)])

    if len(features) != len(y):
        raise Exception(f"Sanity check failed feature list length {len(features)} whereas target list length {len(y)}.")

    sbd_model = create_model()

    scores = cross_val_score(sbd_model, features, y, scoring="f1")

    for i, score in enumerate(scores):
        click.secho(f"Fold {i + 1}: {score:.4f}", fg="yellow")

    sbd_model.fit(features, y)

    click.secho("\nTop 10 Features")
    feature_importance = sbd_model.steps[1][1].feature_importances_
    for idx in list(reversed(feature_importance.argsort()))[:20]:
        click.secho(f"    {sbd_model.steps[0][1].feature_names_[idx]}: {feature_importance[idx]:.4f}", fg="yellow")

    save_model(sbd_model)


if __name__ == '__main__':
    tokenize()
